//
//  Mockingbird.swift
//  BirdMockingbirdExample
//
//  Created by Andrew Chang on 7/26/19.
//

import Foundation
import XCTest

// MARK: - Stubbing

/// Stub mock objects with the same function signature to return a value or perform an operation.
///
/// - Parameter stubbable: A set of stubbable invocations.
public func given<I, R>(_ stubbable: Stubbable<I, R>...) -> Stub<I, R> {
  return Stub<I, R>(from: stubbable)
}

/// Stubs a variable getter to return the last value received by the setter.
///
/// - Parameter initial: The initial value to return.
public func lastSetValue<T, R>(initial: R) -> StubImplementation<T, R> {
  var currentValue = initial
  let implementation: () -> R = { return currentValue }
  let callback = { (stub: StubbingContext.Stub, context: StubbingContext) in
    guard let setterInvocation = stub.invocation.toSetter() else { return }
    let setterImplementation = { (newValue: R) -> Void in
      currentValue = newValue
    }
    _ = context.swizzle(setterInvocation, with: setterImplementation)
  }
  return StubImplementation(handler: implementation as! T, callback: callback)
}

// MARK: - Verification

/// Verify that a mock recieved a specific invocation some number of times.
///
/// - Parameters:
///   - mock: A mockable invocation to verify.
public func verify<T>(file: StaticString = #file, line: UInt = #line,
                      _ mockable: Mockable<T>) -> Verification<T> {
  return Verification(with: mockable, at: SourceLocation(file, line))
}

/// Create a deferrable test expectation from a block containing verification calls.
///
/// - Parameters:
///   - description: An optional description for the created `XCTestExpectation`.
///   - block: A block containing verification calls.
/// - Returns: An XCTestExpectation that fulfilles once all verifications in the block are met.
public func eventually(_ description: String? = nil,
                       _ block: @escaping () -> Void) -> XCTestExpectation {
  return createTestExpectation(with: block, description: description)
}

// MARK: - Expectation resetting

/// Remove all observed invocations _and_ stubbed implementations on a set of mocks.
///
/// - Parameter mocks: A set of mocks to reset.
public func reset<M: Mock>(_ mocks: M...) {
  mocks.forEach({
    $0.mockingContext.clearInvocations()
    $0.stubbingContext.clearStubs()
  })
}

/// Remove all observed invocations on a set of mocks.
///
/// - Parameter mocks: A set of mocks to reset.
public func clearInvocations<M: Mock>(on mocks: M...) {
  mocks.forEach({ $0.mockingContext.clearInvocations() })
}

/// Remove all stubbed implementations on a set of mocks.
///
/// - Parameter mocks: A set of mocks to reset.
public func clearStubs<M: Mock>(on mocks: M...) {
  mocks.forEach({ $0.stubbingContext.clearStubs() })
}

// MARK: - Standard argument matchers

/// Matches any argument value of a specific type `T`.
///
/// - Parameter type: Optionally provide an explicit type to disambiguate overloaded methods.
public func any<T>(_ type: T.Type = T.self) -> T {
  let matcher = ArgumentMatcher(description: "any<\(T.self)>()", priority: .high) { (_, rhs) in
    return rhs is T
  }
  return createTypeFacade(matcher)
}

/// Matches any of the provided values by equality.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - objects: A set of equatable objects that should result in a match.
public func any<T: Equatable>(_ type: T.Type = T.self, of objects: T...) -> T {
  let matcher = ArgumentMatcher(description: "any<\(T.self)>(of: [\(objects)])", priority: .high) {
    (_, rhs) in
    guard let other = rhs as? T else { return false }
    return objects.contains(where: { $0 == other })
  }
  return createTypeFacade(matcher)
}

/// Matches any of the provided values by reference.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - objects: A set of non-equatable objects that should result in a match.
public func any<T>(_ type: T.Type = T.self, of objects: T...) -> T {
  let description = "any<\(T.self)>(of: [\(objects)]) (by reference)"
  let matcher = ArgumentMatcher(description: description, priority: .high) { (_, rhs) in
    return objects.contains(where: { $0 as AnyObject === rhs as AnyObject })
  }
  return createTypeFacade(matcher)
}

/// Matches any values where the predicate returns `true`.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - predicate: A closure that takes a value `T` and returns `true` if it represents a match.
public func any<T>(_ type: T.Type = T.self, where predicate: @escaping (_ value: T) -> Bool) -> T {
  let matcher = ArgumentMatcher(description: "any<\(T.self)>(where:)", priority: .high) {
    (_, rhs) in
    guard let rhs = rhs as? T else { return false }
    return predicate(rhs)
  }
  return createTypeFacade(matcher)
}

/// Matches any non-nil argument value of a specific type `T`.
///
/// - Parameter type: Optionally provide an explicit type to disambiguate overloaded methods.
public func notNil<T>(_ type: T.Type = T.self) -> T {
  let matcher = ArgumentMatcher(description: "notNil<\(T.self)>()", priority: .high) { (_, rhs) in
    return rhs is T && rhs != nil
  }
  return createTypeFacade(matcher)
}

// MARK: Collection matchers

/// Matches any collection containing all of the values.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - values: A set of concrete values to look for in the collection.
public func any<T: Collection>(_ type: T.Type = T.self, containing values: T.Element...) -> T {
  let matcher = ArgumentMatcher(description: "any<\(T.self)>(containing:)", priority: .high) {
    (_, rhs) in
    guard let collection = rhs as? T else { return false }
    return values.allSatisfy({
      let valueMatcher = ArgumentMatcher($0)
      return collection.contains(where: { valueMatcher == ArgumentMatcher($0) })
    })
  }
  return createTypeFacade(matcher)
}

/// Matches any dictionary containing all of the values.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - values: A set of concrete values to look for in the dictionary.
public func any<K, V>(_ type: Dictionary<K, V>.Type = Dictionary<K, V>.self,
                      containing values: V...) -> Dictionary<K, V> {
  let description = "any<\(Dictionary<K, V>.self)>(containing:)"
  let matcher = ArgumentMatcher(description: description, priority: .high) {
    (_, rhs) in
    guard let collection = rhs as? Dictionary<K, V> else { return false }
    return values.allSatisfy({
      let valueMatcher = ArgumentMatcher($0)
      return collection.contains(where: { valueMatcher == ArgumentMatcher($0.value) })
    })
  }
  return createTypeFacade(matcher)
}

/// Matches any dictionary containing all of the keys.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - keys: A set of concrete keys to look for in the dictionary.
public func any<K, V>(_ type: Dictionary<K, V>.Type = Dictionary<K, V>.self,
                      keys: K...) -> Dictionary<K, V> {
  let description = "any<\(Dictionary<K, V>.self)>(keys:)"
  let matcher = ArgumentMatcher(description: description, priority: .high) {
    (_, rhs) in
    guard let collection = rhs as? Dictionary<K, V> else { return false }
    return keys.allSatisfy({
      let keyMatcher = ArgumentMatcher($0)
      return collection.contains(where: { keyMatcher == ArgumentMatcher($0.key) })
    })
  }
  return createTypeFacade(matcher)
}

/// Matches any dictionary containing containing a specific key, which could be a wildcard argument
/// matcher.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - keys: A key to look for in the dictionary, which can be an argument matcher.
public func any<K, V>(_ type: Dictionary<K, V>.Type = Dictionary<K, V>.self,
                      key: @escaping @autoclosure () -> K) -> Dictionary<K, V> {
  let keyMatcher = resolve(key)
  let description = "any<\(Dictionary<K, V>.self)>(keys:)"
  let matcher = ArgumentMatcher(description: description, priority: .high) {
    (_, rhs) in
    guard let collection = rhs as? Dictionary<K, V> else { return false }
    return collection.contains(where: { return keyMatcher == ArgumentMatcher($0.key) })
  }
  return createTypeFacade(matcher)
}

/// Matches any collection containing a specific value, which could be a wildcard argument matcher.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - value: A value to look for within the collection, which can be an argument matcher.
public func any<T: Collection>(_ type: T.Type = T.self,
                               containing value: @escaping @autoclosure () -> T.Element) -> T {
  let valueMatcher = resolve(value)
  let matcher = ArgumentMatcher(description: "any<\(T.self)>(containing:)", priority: .high) {
    (_, rhs) in
    guard let collection = rhs as? T else { return false }
    return collection.contains(where: { return valueMatcher == ArgumentMatcher($0) })
  }
  return createTypeFacade(matcher)
}

/// Matches any collection with a specific number of elements defined by a count matcher.
///
/// - Parameters:
///   - type: Optionally provide an explicit type to disambiguate overloaded methods.
///   - countMatcher: A count matcher defining the number of acceptable elements in the collection.
public func any<T: Collection>(_ type: T.Type = T.self, count countMatcher: CountMatcher) -> T {
  let matcher = ArgumentMatcher(description: "any<\(T.self)>(count:)", priority: .high) {
    (_, rhs) in
    guard let collection = rhs as? T else { return false }
    return countMatcher.matches(UInt(collection.count))
  }
  return createTypeFacade(matcher)
}

/// Matches any collection with at least one element.
///
/// - Parameter type: Optionally provide an explicit type to disambiguate overloaded methods.
public func notEmpty<T: Collection>(_ type: T.Type = T.self) -> T {
  return any(count: atLeast(1))
}

// MARK: - Standard count matchers

/// Matches an exact count.
public func exactly(_ times: UInt) -> CountMatcher {
  return CountMatcher({ $0 == times }, describedBy: { "`\($1)` \($2 ? "≠" : "=") \(times)" })
}

/// Matches a single count.
public var once: CountMatcher { return exactly(1) }

/// Matches a count of zero.
public var never: CountMatcher { return exactly(0) }

/// Matches greater than or equal to some count.
public func atLeast(_ times: UInt) -> CountMatcher {
  return CountMatcher({ $0 >= times }, describedBy: { "`\($1)` \($2 ? "<" : "≥") \(times)" })
}

/// Matches less than or equal to some count.
public func atMost(_ times: UInt) -> CountMatcher {
  return CountMatcher({ $0 <= times }, describedBy: { "`\($1)` \($2 ? ">" : "≤") \(times)" })
}

/// Matches counts that fall within a certain inclusive range.
public func between(_ range: Range<UInt>) -> CountMatcher {
  return atLeast(range.lowerBound).and(atMost(range.upperBound))
}

// MARK: Composing multiple count matchers
extension CountMatcher {
  public func or(_ countMatcher: CountMatcher) -> CountMatcher {
    let matcherCopy = self
    let otherMatcherCopy = countMatcher
    return CountMatcher({ matcherCopy.matcher($0) || otherMatcherCopy.matcher($0) },
      describedBy: {
        let matcherDescription = matcherCopy.describe(invocation: $0, count: $1, negated: $2)
        let otherMatcherDescription = otherMatcherCopy.describe(invocation: $0, count: $1, negated: $2)
        let operand = $2 ? "&&" : "||"
        return "(\(matcherDescription)) \(operand) (\(otherMatcherDescription))"
    })
  }

  public func and(_ countMatcher: CountMatcher) -> CountMatcher {
    let matcherCopy = self
    let otherMatcherCopy = countMatcher
    return CountMatcher({ matcherCopy.matcher($0) && otherMatcherCopy.matcher($0) },
      describedBy: {
        let matcherDescription = matcherCopy.describe(invocation: $0, count: $1, negated: $2)
        let otherMatcherDescription = otherMatcherCopy.describe(invocation: $0, count: $1, negated: $2)
        let operand = $2 ? "||" : "&&"
        return "(\(matcherDescription)) \(operand) (\(otherMatcherDescription))"
    })
  }

  public func xor(_ countMatcher: CountMatcher) -> CountMatcher {
    let matcherCopy = self
    let otherMatcherCopy = countMatcher
    return CountMatcher({ matcherCopy.matcher($0) != otherMatcherCopy.matcher($0) },
      describedBy: {
        let matcherDescription = matcherCopy.describe(invocation: $0, count: $1, negated: $2)
        let otherMatcherDescription = otherMatcherCopy.describe(invocation: $0, count: $1, negated: $2)
        let operand = $2 ? "≠" : "="
        return "(\(matcherDescription)) \(operand) (\(otherMatcherDescription))"
    })
  }
}

public func not(_ countMatcher: CountMatcher) -> CountMatcher {
  let matcherCopy = countMatcher
  return CountMatcher({ !matcherCopy.matcher($0) },
    describedBy: {
      let matcherDescription = matcherCopy.describe(invocation: $0, count: $1, negated: !$2)
      return "\(matcherDescription)"
  })
}
